@@ -1,14 +1,15 @@
 import logging
+import re
 import string
 from logging import Logger
-from typing import Optional, List, Set
+from typing import Optional, List, Set, Any
 
 from BaseClasses import Item, ItemClassification, Location, MultiWorld, Region, Tutorial
 from Options import OptionError
-
-from .Characters import character_list, CharacterConfig, character_option_map, character_offset_map, NUM_CUSTOM
+from .Characters import character_list, CharacterConfig, character_offset_map, NUM_CUSTOM
 from .Items import event_item_pairs, item_table, ItemType, chars_to_items, base_event_item_pairs, item_groups
-from .Locations import location_table, loc_ids_to_data, LocationData, LocationType, CARD_DRAW_COUNT, location_groups
+from .Locations import location_table, loc_ids_to_data, LocationData, LocationType, CARD_REWARD_COUNT, location_groups, \
+    CHAR_OFFSET
 from .Options import SpireOptions, option_groups
 from .Regions import create_regions
 from .Rules import set_rules
@@ -46,6 +47,8 @@ class SpireWorld(World):
     location_name_groups = location_groups
     item_name_groups = item_groups
 
+    ut_can_gen_without_yaml = True
+
     item_name_to_id = {name: data.code for name, data in item_table.items()}
     location_name_to_id = location_table
     logger = logging.getLogger("SlayTheSpire")
@@ -59,72 +62,22 @@ class SpireWorld(World):
         self.total_shop_items = 0
 
     def generate_early(self):
+        if hasattr(self.multiworld, 're_gen_passthrough'):
+            self._setup_ut(self.multiworld.re_gen_passthrough[self.game])
+            return
         if self.options.use_advanced_characters.value == 0:
-            char_options = self.options.character.value
-            num_rand_chars = self.options.pick_num_characters.value
-            if num_rand_chars != 0 and num_rand_chars < len(char_options):
-                char_options = self.random.sample(list(char_options), k=num_rand_chars)
-            unlocked_char = self._get_unlocked_char(char_options)
-            self.logger.info("Generating with characters %s", char_options)
-            for char_val in char_options:
-                option_name = char_val
-                char_offset = character_offset_map[option_name.lower()]
-                name = character_list[char_offset]
-                if self.options.seeded:
-                    seed = "".join(self.random.choice(string.ascii_letters) for i in range(16))
-                else:
-                    seed = ""
-                locked = False if unlocked_char is None or unlocked_char.lower() == option_name.lower() else True
-
-                config = CharacterConfig(name,
-                                         option_name,
-                                         char_offset,
-                                         0,
-                                         seed,
-                                         locked,
-                                         ascension=self.options.ascension.value,
-                                         final_act=self.options.final_act.value==1,
-                                         downfall=self.options.downfall.value==1)
-                self.characters.append(config)
+            self._handle_basic_chars()
         else:
-            advanced_chars = self.options.advanced_characters.keys()
-            num_rand_chars = self.options.pick_num_characters.value
-            if num_rand_chars != 0 and num_rand_chars < len(advanced_chars):
-                advanced_chars = self.random.sample(list(advanced_chars), k=num_rand_chars)
-            unlocked_char = self._get_unlocked_char(advanced_chars)
-            self.logger.info("Generating with characters %s", advanced_chars)
-            for option_name in advanced_chars:
-                options = self.options.advanced_characters[option_name]
-                mod_num = 0
-                char_offset = character_offset_map.get(option_name.lower(), None)
-                if char_offset is None:
-                    self.modded_num += 1
-                    mod_num = self.modded_num
-                    char_offset = mod_num + len(character_list) - 1
-                    name = f"Custom Character {mod_num}"
-                else:
-                    name = character_list[char_offset]
-                if self.options.seeded:
-                    seed = "".join(self.random.choice(string.ascii_letters) for i in range(16))
-                else:
-                    seed = ""
-                locked = False if unlocked_char is None or unlocked_char.lower() == option_name.lower() else True
-                config = CharacterConfig(name,
-                                         option_name,
-                                         char_offset,
-                                         mod_num,
-                                         seed,
-                                         locked,
-                                         **options)
-                self.characters.append(config)
-                if config.mod_num > 0:
-                    self.modded_chars.append(config)
+            self._handle_advanced_chars()
+
+        if not self.characters:
+            raise OptionError("At least one character must be configured")
         names = set()
         for config in self.characters:
             self.logger.info("StS: Got character configuration" + str(config))
             names.add(config.official_name)
         if len(names) != len(self.characters):
-            raise OptionError(f"Found duplicate characters: {names}")
+            raise OptionError(f"Found duplicate characters: {[x.official_name for x in self.characters]}")
         for config in self.characters:
             if not config.locked:
                 break
@@ -141,8 +94,13 @@ class SpireWorld(World):
         if num_chars_goal != 0:
             if num_chars_goal > len(self.characters):
                 self.options.num_chars_goal.value = 0
+        for weight in self.options.trap_weights.values():
+            if weight > 0:
+                break
+        else:
+            self.options.trap_chance.value = 0
 
-    def _get_unlocked_char(self, characters: Set[str]) -> Optional[str]:
+    def _get_unlocked_char(self, characters: List[str]) -> Optional[str]:
         if len(characters) <= 0:
             raise OptionError("At least one character must be selected.")
         locked_opt = self.options.lock_characters.value
@@ -151,21 +109,118 @@ class SpireWorld(World):
             unlocked_char = self.random.choice([x for x in characters])
         elif locked_opt == 2:
             unlocked_char = self.options.unlocked_character.value
-            if unlocked_char not in characters:
+            if type(unlocked_char) == int:
+                unlocked_char = character_list[unlocked_char]
+
+            for char in characters:
+                if char.lower() == unlocked_char.lower():
+                    return char
+            else:
                 raise OptionError(
                     f"Configured {unlocked_char} as the first unlocked character, but was not one of: {characters}")
         return unlocked_char
 
+    def _handle_basic_chars(self) -> None:
+        if len(self.options.character.value) > 0:
+            self.logger.warning("The 'character' option has been renamed to 'characters'; please update your yaml")
+            self.options.characters.value = self.options.character.value
+        selected_chars = sorted(self.options.characters.value)
+        num_rand_chars = self.options.pick_num_characters.value
+        unlocked_char = self._get_unlocked_char(selected_chars)
+        if self.options.lock_characters.value != 0 and num_rand_chars != 0 and num_rand_chars < len(selected_chars):
+            selected_chars.remove(unlocked_char)
+            selected_chars = [unlocked_char] + self.random.sample(selected_chars, k=num_rand_chars - 1)
+        self.logger.info("Generating with characters %s", selected_chars)
+        ascension_down = self.options.ascension_down.value
+        if self.options.include_floor_checks.value == 0:
+            ascension_down = 0
+        for char_val in selected_chars:
+            option_name = char_val
+            char_offset = character_offset_map[option_name.lower()]
+            name = character_list[char_offset]
+            if self.options.seeded:
+                seed = "".join(self.random.choice(string.ascii_letters) for i in range(16))
+            else:
+                seed = ""
+            locked = False if unlocked_char is None or unlocked_char.lower() == option_name.lower() else True
+            config = CharacterConfig(name,
+                                     option_name,
+                                     char_offset,
+                                     0,
+                                     seed,
+                                     locked,
+                                     ascension=self.options.ascension.value,
+                                     final_act=self.options.final_act.value == 1,
+                                     downfall=self.options.downfall.value == 1,
+                                     ascension_down=ascension_down)
+            self.characters.append(config)
+
+    def _handle_advanced_chars(self) -> None:
+        advanced_chars = self.options.advanced_characters.keys()
+        char_options = sorted(advanced_chars)
+        num_rand_chars = self.options.pick_num_characters.value
+        unlocked_char = self._get_unlocked_char(char_options)
+        include_ascension_down = self.options.include_floor_checks.value != 0
+        if self.options.lock_characters.value != 0 and num_rand_chars != 0 and num_rand_chars < len(char_options):
+            selected_chars = list(char_options)
+            if unlocked_char in selected_chars:
+                selected_chars.remove(unlocked_char)
+            selected_chars = [unlocked_char] + self.random.sample(selected_chars, k=num_rand_chars - 1)
+            modded_num = 0
+            for char in selected_chars:
+                if character_offset_map.get(char.lower(), None) is None:
+                    modded_num += 1
+            if modded_num > NUM_CUSTOM:
+                supported_chars = sorted({x for x in char_options if x.lower() in character_offset_map})
+                replace_num = modded_num - NUM_CUSTOM
+                remove_me = self.random.sample(selected_chars, k=replace_num)
+                for remove in remove_me:
+                    selected_chars.remove(remove)
+                selected_chars += self.random.sample(supported_chars, k=min(replace_num, len(supported_chars)))
+        else:
+            selected_chars = char_options
+
+        self.logger.info("Generating with characters %s", selected_chars)
+        for option_name in selected_chars:
+            options = self.options.advanced_characters[option_name]
+            mod_num = 0
+            char_offset = character_offset_map.get(option_name.lower(), None)
+            if char_offset is None:
+                self.modded_num += 1
+                mod_num = self.modded_num
+                char_offset = mod_num + len(character_list) - 1
+                name = f"Custom Character {mod_num}"
+            else:
+                name = character_list[char_offset]
+            if self.options.seeded:
+                seed = "".join(self.random.choice(string.ascii_letters) for i in range(16))
+            else:
+                seed = ""
+            locked = False if unlocked_char is None or unlocked_char.lower() == option_name.lower() else True
+            config = CharacterConfig(name,
+                                     option_name,
+                                     char_offset,
+                                     mod_num,
+                                     seed,
+                                     locked,
+                                     **options)
+            if not include_ascension_down:
+                config.ascension_down = 0
+            self.characters.append(config)
+            if config.mod_num > 0:
+                self.modded_chars.append(config)
+
     def create_items(self):
         # Fill out our pool with our items from item_pool, assuming 1 item if not present in item_pool
         pool = []
         for config in self.characters:
             char_lookup = config.name if config.mod_num == 0 else config.mod_num
+            ascension_downs = min(config.ascension_down, config.ascension)
             for name, data in chars_to_items[char_lookup].items():
                 amount = 0
-                if ItemType.DRAW == data.type:
-                    amount = CARD_DRAW_COUNT
-                elif ItemType.RARE_DRAW == data.type or ItemType.BOSS_RELIC == data.type:
+                if ItemType.CARD_REWARD == data.type:
+                    amount = CARD_REWARD_COUNT
+                elif ItemType.RARE_CARD_REWARD == data.type or ItemType.BOSS_RELIC == data.type:
                     amount = 2
                 elif ItemType.RELIC == data.type:
                     amount = 10
@@ -182,6 +237,8 @@ class SpireWorld(World):
                         amount = 2
                 elif ItemType.POTION == data.type and self.options.potion_sanity:
                     amount = 9
+                elif ItemType.ASCENSION_DOWN == data.type and self.options.include_floor_checks.value != 0:
+                    amount = ascension_downs
                 elif self.options.shop_sanity.value != 0:
                     if ItemType.SHOP_CARD == data.type:
                         amount = self.options.shop_card_slots.value
@@ -197,14 +254,21 @@ class SpireWorld(World):
                     pool.append(SpireItem(name, self.player))
 
             if self.options.include_floor_checks.value:
-                remaining_checks = 51
+
+                remaining_checks = 51 - ascension_downs
 
                 if config.final_act:
                     remaining_checks += 4
                 if config.ascension >= 20:
                     remaining_checks += 1
+
+                traps: list[bool] = [self.random.randint(0, 100) < self.options.trap_chance for _ in range(remaining_checks)]
+                trap_num = traps.count(True)
+                filler_num = len(traps) - trap_num
+                for name in self.random.choices(list(self.options.trap_weights.keys()), weights=list(self.options.trap_weights.values()),k=trap_num):
+                    pool.append(SpireItem(name, self.player))
                 for name in self.random.choices([key for key, val in chars_to_items[char_lookup].items()
-                                                 if ItemType.GOLD == val.type and ItemClassification.filler == val.classification], weights=[40,60],k=remaining_checks):
+                                                 if ItemType.GOLD == val.type and ItemClassification.filler == val.classification], weights=[40,60],k=filler_num):
                     pool.append(SpireItem(name, self.player))
             # Pair up our event locations with our event items
             for base_event, base_item in base_event_item_pairs.items():
@@ -240,7 +304,6 @@ class SpireWorld(World):
             "mod_version": self.mod_version,
         }
         slot_data.update(self.options.as_dict(
-            "character",
             "ascension",
             "final_act",
             "downfall",
@@ -256,10 +319,11 @@ class SpireWorld(World):
         return slot_data
 
     def get_filler_item_name(self) -> str:
-        config = self.characters[0]
+        if not self.characters:
+            return "CAW CAW"
+        config = self.random.choice(self.characters)
         return self.random.choice([f"{config.name} One Gold", f"{config.name} Five Gold"])
 
-
     def create_region(self, player: int, prefix: Optional[str], name: str, config: CharacterConfig, locations: List[str] = None, exits: List[str] =None):
         ret = Region(f"{prefix} {name}" if prefix is not None else name, player, self.multiworld)
         if locations:
@@ -297,6 +361,60 @@ class SpireWorld(World):
             return False
         return True
 
+    @staticmethod
+    def interpret_slot_data(slot_data: dict[str, Any]) -> Any:
+        return slot_data
+
+    def _setup_ut(self, slot_data: dict[str, Any]) -> None:
+        self.options.shop_remove_slots.value = slot_data["shop_sanity_options"]["card_remove"]
+        self.options.shop_neutral_card_slots.value = slot_data["shop_sanity_options"]["neutral_slots"]
+        self.options.shop_relic_slots.value = slot_data["shop_sanity_options"]["relic_slots"]
+        self.options.shop_potion_slots.value = slot_data["shop_sanity_options"]["potion_slots"]
+        for char_dict in slot_data['characters']:
+            config = CharacterConfig(
+                char_dict['name'],
+                char_dict['option_name'],
+                char_dict['char_offset'],
+                char_dict['mod_num'],
+                char_dict['seed'],
+                char_dict['locked'],
+                ascension=char_dict['ascension'],
+                final_act=char_dict['final_act'],
+                downfall=char_dict['downfall'],
+                ascension_down=char_dict['ascension_down'],
+            )
+            self.characters.append(config)
+            if char_dict['mod_num'] > 0:
+                self.modded_chars.append(config)
+        self.total_shop_items = (self.options.shop_card_slots.value + self.options.shop_neutral_card_slots.value +
+                                 self.options.shop_relic_slots.value + self.options.shop_potion_slots.value)
+        self.total_shop_locations = self.total_shop_items + (3 if self.options.shop_remove_slots else 0)
+        if self.total_shop_locations <= 0:
+            self.options.shop_sanity.value = 0
+        self.options.include_floor_checks.value = slot_data['include_floor_checks']
+        self.options.campfire_sanity.value = slot_data['campfire_sanity']
+        self.options.shop_sanity.value = slot_data['shop_sanity']
+        self.options.gold_sanity.value = slot_data['gold_sanity']
+        self.options.potion_sanity.value = slot_data['potion_sanity']
+        self.options.num_chars_goal.value = slot_data['num_chars_goal']
+        self.location_id_to_alias: dict[int, str] = dict()
+        pattern = re.compile("Custom Character [0-9]+ (?P<location_name>.*?)$")
+        # for i in range(1, len(self.modded_chars) + 1):
+        for key, value in SpireWorld.location_id_to_name.items():
+            if key < (len(character_list)) * CHAR_OFFSET:
+                continue
+            modded_index = (key // CHAR_OFFSET) - len(character_list)
+            self.logger.info(f"Modded index: {modded_index}")
+            self.logger.info(f"modded_chars index: {self.modded_chars}")
+            if modded_index >= len(self.modded_chars):
+                continue
+            match = pattern.match(value)
+            if match is None:
+                raise Exception("Failed to match " + value)
+            name = self.modded_chars[modded_index].official_name
+            self.logger.info(name)
+            self.location_id_to_alias[key] = name + " " + match.group("location_name")
+
 
 class SpireLocation(Location):
     game: str = "Slay the Spire"
